@@ -3,7 +3,7 @@
 #     A Python interface to Last.fm and Libre.fm
 #
 # Copyright 2008-2010 Amr Hassan
-# Copyright 2013-2020 hugovk
+# Copyright 2013-2021 hugovk
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -23,6 +23,7 @@ import collections
 import hashlib
 import html.entities
 import logging
+import os
 import shelve
 import ssl
 import tempfile
@@ -35,7 +36,7 @@ from xml.dom import Node, minidom
 import pkg_resources
 
 __author__ = "Amr Hassan, hugovk, Mice Pápai"
-__copyright__ = "Copyright (C) 2008-2010 Amr Hassan, 2013-2020 hugovk, 2017 Mice Pápai"
+__copyright__ = "Copyright (C) 2008-2010 Amr Hassan, 2013-2021 hugovk, 2017 Mice Pápai"
 __license__ = "apache2"
 __email__ = "amr.hassan@gmail.com"
 __version__ = pkg_resources.get_distribution(__name__).version
@@ -421,7 +422,8 @@ class _Network:
         """
 
         if not file_path:
-            file_path = tempfile.mktemp(prefix="pylast_tmp_")
+            self.cache_backend = _ShelfCacheBackend.create_shelf()
+            return
 
         self.cache_backend = _ShelfCacheBackend(file_path)
 
@@ -665,8 +667,7 @@ class LastFMNetwork(_Network):
         password_hash="",
         token="",
     ):
-        _Network.__init__(
-            self,
+        super().__init__(
             name="Last.fm",
             homepage="https://www.last.fm",
             ws_server=("ws.audioscrobbler.com", "/2.0/"),
@@ -733,8 +734,7 @@ class LibreFMNetwork(_Network):
         self, api_key="", api_secret="", session_key="", username="", password_hash=""
     ):
 
-        _Network.__init__(
-            self,
+        super().__init__(
             name="Libre.fm",
             homepage="https://libre.fm",
             ws_server=("libre.fm", "/2.0/"),
@@ -784,8 +784,11 @@ class LibreFMNetwork(_Network):
 class _ShelfCacheBackend:
     """Used as a backend for caching cacheable requests."""
 
-    def __init__(self, file_path=None):
-        self.shelf = shelve.open(file_path)
+    def __init__(self, file_path=None, flag=None):
+        if flag is not None:
+            self.shelf = shelve.open(file_path, flag=flag)
+        else:
+            self.shelf = shelve.open(file_path)
         self.cache_keys = set(self.shelf.keys())
 
     def __contains__(self, key):
@@ -801,6 +804,12 @@ class _ShelfCacheBackend:
         self.cache_keys.add(key)
         self.shelf[key] = xml_string
 
+    @classmethod
+    def create_shelf(cls):
+        file_descriptor, file_path = tempfile.mkstemp(prefix="pylast_tmp_")
+        os.close(file_descriptor)
+        return cls(file_path=file_path, flag="n")
+
 
 class _Request:
     """Representing an abstract web service operation."""
@@ -918,7 +927,7 @@ class _Request:
                     headers=headers,
                 )
             except Exception as e:
-                raise NetworkError(self.network, e)
+                raise NetworkError(self.network, e) from e
 
         else:
             conn = HTTPSConnection(context=SSL_CONTEXT, host=host_name)
@@ -926,7 +935,7 @@ class _Request:
             try:
                 conn.request(method="POST", url=host_subdir, body=data, headers=headers)
             except Exception as e:
-                raise NetworkError(self.network, e)
+                raise NetworkError(self.network, e) from e
 
         try:
             response = conn.getresponse()
@@ -939,7 +948,7 @@ class _Request:
                 )
             response_text = _unicode(response.read())
         except Exception as e:
-            raise MalformedResponseError(self.network, e)
+            raise MalformedResponseError(self.network, e) from e
 
         try:
             self._check_response_for_errors(response_text)
@@ -963,7 +972,7 @@ class _Request:
         try:
             doc = minidom.parseString(_string(response).replace("opensearch:", ""))
         except Exception as e:
-            raise MalformedResponseError(self.network, e)
+            raise MalformedResponseError(self.network, e) from e
 
         e = doc.getElementsByTagName("lfm")[0]
         # logger.debug(doc.toprettyxml())
@@ -1044,9 +1053,6 @@ class SessionKeyGenerator:
 
         if url in self.web_auth_tokens.keys():
             token = self.web_auth_tokens[url]
-        else:
-            # This will raise a WSError if token is blank or unauthorized
-            token = token
 
         request = _Request(self.network, "auth.getSession", {"token": token})
 
@@ -1146,21 +1152,29 @@ class _BaseObject:
 
         return first_child.wholeText.strip()
 
-    def _get_things(self, method, thing, thing_type, params=None, cacheable=True):
+    def _get_things(
+        self, method, thing_type, params=None, cacheable=True, stream=False
+    ):
         """Returns a list of the most played thing_types by this thing."""
 
-        limit = params.get("limit", 1)
-        seq = []
-        for node in _collect_nodes(
-            limit, self, self.ws_prefix + "." + method, cacheable, params
-        ):
-            title = _extract(node, "name")
-            artist = _extract(node, "name", 1)
-            playcount = _number(_extract(node, "playcount"))
+        def _stream_get_things():
+            limit = params.get("limit", 50)
+            nodes = _collect_nodes(
+                limit,
+                self,
+                self.ws_prefix + "." + method,
+                cacheable,
+                params,
+                stream=stream,
+            )
+            for node in nodes:
+                title = _extract(node, "name")
+                artist = _extract(node, "name", 1)
+                playcount = _number(_extract(node, "playcount"))
 
-            seq.append(TopItem(thing_type(artist, title, self.network), playcount))
+                yield TopItem(thing_type(artist, title, self.network), playcount)
 
-        return seq
+        return _stream_get_things() if stream else list(_stream_get_things())
 
     def get_wiki_published_date(self):
         """
@@ -1201,11 +1215,11 @@ class _BaseObject:
         return _extract(node, section)
 
 
-class _Chartable:
+class _Chartable(_BaseObject):
     """Common functions for classes with charts."""
 
-    def __init__(self, ws_prefix):
-        self.ws_prefix = ws_prefix  # TODO move to _BaseObject?
+    def __init__(self, network, ws_prefix):
+        super().__init__(network=network, ws_prefix=ws_prefix)
 
     def get_weekly_chart_dates(self):
         """Returns a list of From and To tuples for the available charts."""
@@ -1272,11 +1286,11 @@ class _Chartable:
         return seq
 
 
-class _Taggable:
+class _Taggable(_BaseObject):
     """Common functions for classes with tags."""
 
-    def __init__(self, ws_prefix):
-        self.ws_prefix = ws_prefix  # TODO move to _BaseObject
+    def __init__(self, network, ws_prefix):
+        super().__init__(network=network, ws_prefix=ws_prefix)
 
     def add_tags(self, tags):
         """Adds one or several tags.
@@ -1381,9 +1395,9 @@ class _Taggable:
 
         for element in elements:
             tag_name = _extract(element, "name")
-            tagcount = _extract(element, "count")
+            tag_count = _extract(element, "count")
 
-            seq.append(TopItem(Tag(tag_name, self.network), tagcount))
+            seq.append(TopItem(Tag(tag_name, self.network), tag_count))
 
         if limit:
             seq = seq[:limit]
@@ -1391,7 +1405,13 @@ class _Taggable:
         return seq
 
 
-class WSError(Exception):
+class PyLastError(Exception):
+    """Generic exception raised by PyLast"""
+
+    pass
+
+
+class WSError(PyLastError):
     """Exception related to the Network web service"""
 
     def __init__(self, network, status, details):
@@ -1435,7 +1455,7 @@ class WSError(Exception):
         return self.status
 
 
-class MalformedResponseError(Exception):
+class MalformedResponseError(PyLastError):
     """Exception conveying a malformed response from the music network."""
 
     def __init__(self, network, underlying_error):
@@ -1448,7 +1468,7 @@ class MalformedResponseError(Exception):
         )
 
 
-class NetworkError(Exception):
+class NetworkError(PyLastError):
     """Exception conveying a problem in sending a request to Last.fm"""
 
     def __init__(self, network, underlying_error):
@@ -1459,7 +1479,7 @@ class NetworkError(Exception):
         return "NetworkError: %s" % str(self.underlying_error)
 
 
-class _Opus(_BaseObject, _Taggable):
+class _Opus(_Taggable):
     """An album or track."""
 
     artist = None
@@ -1480,8 +1500,7 @@ class _Opus(_BaseObject, _Taggable):
         if info is None:
             info = {}
 
-        _BaseObject.__init__(self, network, ws_prefix)
-        _Taggable.__init__(self, ws_prefix)
+        super().__init__(network=network, ws_prefix=ws_prefix)
 
         if isinstance(artist, Artist):
             self.artist = artist
@@ -1649,7 +1668,7 @@ class Album(_Opus):
         }
 
 
-class Artist(_BaseObject, _Taggable):
+class Artist(_Taggable):
     """An artist."""
 
     name = None
@@ -1666,8 +1685,7 @@ class Artist(_BaseObject, _Taggable):
         if info is None:
             info = {}
 
-        _BaseObject.__init__(self, network, "artist")
-        _Taggable.__init__(self, "artist")
+        super().__init__(network=network, ws_prefix="artist")
 
         self.name = name
         self.username = username
@@ -1806,21 +1824,21 @@ class Artist(_BaseObject, _Taggable):
 
         return artists
 
-    def get_top_albums(self, limit=None, cacheable=True):
+    def get_top_albums(self, limit=None, cacheable=True, stream=False):
         """Returns a list of the top albums."""
         params = self._get_params()
         if limit:
             params["limit"] = limit
 
-        return self._get_things("getTopAlbums", "album", Album, params, cacheable)
+        return self._get_things("getTopAlbums", Album, params, cacheable, stream=stream)
 
-    def get_top_tracks(self, limit=None, cacheable=True):
+    def get_top_tracks(self, limit=None, cacheable=True, stream=False):
         """Returns a list of the most played Tracks by this artist."""
         params = self._get_params()
         if limit:
             params["limit"] = limit
 
-        return self._get_things("getTopTracks", "track", Track, params, cacheable)
+        return self._get_things("getTopTracks", Track, params, cacheable, stream=stream)
 
     def get_url(self, domain_name=DOMAIN_ENGLISH):
         """Returns the URL of the artist page on the network.
@@ -1853,7 +1871,7 @@ class Country(_BaseObject):
     __hash__ = _BaseObject.__hash__
 
     def __init__(self, name, network):
-        _BaseObject.__init__(self, network, "geo")
+        super().__init__(network=network, ws_prefix="geo")
 
         self.name = name
 
@@ -1888,13 +1906,13 @@ class Country(_BaseObject):
 
         return _extract_top_artists(doc, self)
 
-    def get_top_tracks(self, limit=None, cacheable=True):
+    def get_top_tracks(self, limit=None, cacheable=True, stream=False):
         """Returns a sequence of the most played tracks"""
         params = self._get_params()
         if limit:
             params["limit"] = limit
 
-        return self._get_things("getTopTracks", "track", Track, params, cacheable)
+        return self._get_things("getTopTracks", Track, params, cacheable, stream=stream)
 
     def get_url(self, domain_name=DOMAIN_ENGLISH):
         """Returns the URL of the country page on the network.
@@ -1928,7 +1946,7 @@ class Library(_BaseObject):
     __hash__ = _BaseObject.__hash__
 
     def __init__(self, user, network):
-        _BaseObject.__init__(self, network, "library")
+        super().__init__(network=network, ws_prefix="library")
 
         if isinstance(user, User):
             self.user = user
@@ -1949,27 +1967,27 @@ class Library(_BaseObject):
         """Returns the user who owns this library."""
         return self.user
 
-    def get_artists(self, limit=50, cacheable=True):
+    def get_artists(self, limit=50, cacheable=True, stream=False):
         """
         Returns a sequence of Album objects
         if limit==None it will return all (may take a while)
         """
 
-        seq = []
-        for node in _collect_nodes(
-            limit, self, self.ws_prefix + ".getArtists", cacheable
-        ):
-            name = _extract(node, "name")
+        def _get_artists():
+            for node in _collect_nodes(
+                limit, self, self.ws_prefix + ".getArtists", cacheable, stream=stream
+            ):
+                name = _extract(node, "name")
 
-            playcount = _number(_extract(node, "playcount"))
-            tagcount = _number(_extract(node, "tagcount"))
+                playcount = _number(_extract(node, "playcount"))
+                tagcount = _number(_extract(node, "tagcount"))
 
-            seq.append(LibraryItem(Artist(name, self.network), playcount, tagcount))
+                yield LibraryItem(Artist(name, self.network), playcount, tagcount)
 
-        return seq
+        return _get_artists() if stream else list(_get_artists())
 
 
-class Tag(_BaseObject, _Chartable):
+class Tag(_Chartable):
     """A Last.fm object tag."""
 
     name = None
@@ -1977,8 +1995,7 @@ class Tag(_BaseObject, _Chartable):
     __hash__ = _BaseObject.__hash__
 
     def __init__(self, name, network):
-        _BaseObject.__init__(self, network, "tag")
-        _Chartable.__init__(self, "tag")
+        super().__init__(network=network, ws_prefix="tag")
 
         self.name = name
 
@@ -2018,13 +2035,13 @@ class Tag(_BaseObject, _Chartable):
 
         return _extract_top_albums(doc, self.network)
 
-    def get_top_tracks(self, limit=None, cacheable=True):
+    def get_top_tracks(self, limit=None, cacheable=True, stream=False):
         """Returns a list of the most played Tracks for this tag."""
         params = self._get_params()
         if limit:
             params["limit"] = limit
 
-        return self._get_things("getTopTracks", "track", Track, params, cacheable)
+        return self._get_things("getTopTracks", Track, params, cacheable, stream=stream)
 
     def get_top_artists(self, limit=None, cacheable=True):
         """Returns a sequence of the most played artists."""
@@ -2180,7 +2197,7 @@ class Track(_Opus):
         }
 
 
-class User(_BaseObject, _Chartable):
+class User(_Chartable):
     """A Last.fm user."""
 
     name = None
@@ -2188,8 +2205,7 @@ class User(_BaseObject, _Chartable):
     __hash__ = _BaseObject.__hash__
 
     def __init__(self, user_name, network):
-        _BaseObject.__init__(self, network, "user")
-        _Chartable.__init__(self, "user")
+        super().__init__(network=network, ws_prefix="user")
 
         self.name = user_name
 
@@ -2212,6 +2228,16 @@ class User(_BaseObject, _Chartable):
     def _get_params(self):
         return {self.ws_prefix: self.get_name()}
 
+    def _extract_played_track(self, track_node):
+        title = _extract(track_node, "name")
+        track_artist = _extract(track_node, "artist")
+        date = _extract(track_node, "date")
+        album = _extract(track_node, "album")
+        timestamp = track_node.getElementsByTagName("date")[0].getAttribute("uts")
+        return PlayedTrack(
+            Track(track_artist, title, self.network), album, date, timestamp
+        )
+
     def get_name(self, properly_capitalized=False):
         """Returns the user name."""
 
@@ -2222,47 +2248,53 @@ class User(_BaseObject, _Chartable):
 
         return self.name
 
-    def get_friends(self, limit=50, cacheable=False):
+    def get_friends(self, limit=50, cacheable=False, stream=False):
         """Returns a list of the user's friends. """
 
-        seq = []
-        for node in _collect_nodes(
-            limit, self, self.ws_prefix + ".getFriends", cacheable
-        ):
-            seq.append(User(_extract(node, "name"), self.network))
+        def _get_friends():
+            for node in _collect_nodes(
+                limit, self, self.ws_prefix + ".getFriends", cacheable, stream=stream
+            ):
+                yield User(_extract(node, "name"), self.network)
 
-        return seq
+        return _get_friends() if stream else list(_get_friends())
 
-    def get_loved_tracks(self, limit=50, cacheable=True):
+    def get_loved_tracks(self, limit=50, cacheable=True, stream=False):
         """
         Returns this user's loved track as a sequence of LovedTrack objects in
         reverse order of their timestamp, all the way back to the first track.
 
         If limit==None, it will try to pull all the available data.
+        If stream=True, it will yield tracks as soon as a page has been retrieved.
 
         This method uses caching. Enable caching only if you're pulling a
         large amount of data.
         """
 
-        params = self._get_params()
-        if limit:
-            params["limit"] = limit
+        def _get_loved_tracks():
+            params = self._get_params()
+            if limit:
+                params["limit"] = limit
 
-        seq = []
-        for track in _collect_nodes(
-            limit, self, self.ws_prefix + ".getLovedTracks", cacheable, params
-        ):
-            try:
-                artist = _extract(track, "name", 1)
-            except IndexError:  # pragma: no cover
-                continue
-            title = _extract(track, "name")
-            date = _extract(track, "date")
-            timestamp = track.getElementsByTagName("date")[0].getAttribute("uts")
+            for track in _collect_nodes(
+                limit,
+                self,
+                self.ws_prefix + ".getLovedTracks",
+                cacheable,
+                params,
+                stream=stream,
+            ):
+                try:
+                    artist = _extract(track, "name", 1)
+                except IndexError:  # pragma: no cover
+                    continue
+                title = _extract(track, "name")
+                date = _extract(track, "date")
+                timestamp = track.getElementsByTagName("date")[0].getAttribute("uts")
 
-            seq.append(LovedTrack(Track(artist, title, self.network), date, timestamp))
+                yield LovedTrack(Track(artist, title, self.network), date, timestamp)
 
-        return seq
+        return _get_loved_tracks() if stream else list(_get_loved_tracks())
 
     def get_now_playing(self):
         """
@@ -2290,7 +2322,15 @@ class User(_BaseObject, _Chartable):
 
         return Track(artist, title, self.network, self.name, info=info)
 
-    def get_recent_tracks(self, limit=10, cacheable=True, time_from=None, time_to=None):
+    def get_recent_tracks(
+        self,
+        limit=10,
+        cacheable=True,
+        time_from=None,
+        time_to=None,
+        stream=False,
+        now_playing=False,
+    ):
         """
         Returns this user's played track as a sequence of PlayedTrack objects
         in reverse order of playtime, all the way back to the first track.
@@ -2305,45 +2345,39 @@ class User(_BaseObject, _Chartable):
         before this time, in UNIX timestamp format (integer number of
         seconds since 00:00:00, January 1st 1970 UTC). This must be in
         the UTC time zone.
+        stream: If True, it will yield tracks as soon as a page has been retrieved.
 
         This method uses caching. Enable caching only if you're pulling a
         large amount of data.
         """
 
-        params = self._get_params()
-        if limit:
-            params["limit"] = limit + 1  # in case we remove the now playing track
-        if time_from:
-            params["from"] = time_from
-        if time_to:
-            params["to"] = time_to
+        def _get_recent_tracks():
+            params = self._get_params()
+            if limit:
+                params["limit"] = limit + 1  # in case we remove the now playing track
+            if time_from:
+                params["from"] = time_from
+            if time_to:
+                params["to"] = time_to
 
-        seq = []
-        for track in _collect_nodes(
-            limit + 1 if limit else None,
-            self,
-            self.ws_prefix + ".getRecentTracks",
-            cacheable,
-            params,
-        ):
+            track_count = 0
+            for track_node in _collect_nodes(
+                limit + 1 if limit else None,
+                self,
+                self.ws_prefix + ".getRecentTracks",
+                cacheable,
+                params,
+                stream=stream,
+            ):
+                if track_node.hasAttribute("nowplaying") and not now_playing:
+                    continue  # to prevent the now playing track from sneaking in
 
-            if track.hasAttribute("nowplaying"):
-                continue  # to prevent the now playing track from sneaking in
+                if limit and track_count >= limit:
+                    break
+                yield self._extract_played_track(track_node=track_node)
+                track_count += 1
 
-            title = _extract(track, "name")
-            artist = _extract(track, "artist")
-            date = _extract(track, "date")
-            album = _extract(track, "album")
-            timestamp = track.getElementsByTagName("date")[0].getAttribute("uts")
-
-            seq.append(
-                PlayedTrack(Track(artist, title, self.network), album, date, timestamp)
-            )
-
-        if limit:
-            # Slice, in case we didn't remove a now playing track
-            seq = seq[:limit]
-        return seq
+        return _get_recent_tracks() if stream else list(_get_recent_tracks())
 
     def get_country(self):
         """Returns the name of the country of the user."""
@@ -2482,7 +2516,9 @@ class User(_BaseObject, _Chartable):
 
         return seq
 
-    def get_top_tracks(self, period=PERIOD_OVERALL, limit=None, cacheable=True):
+    def get_top_tracks(
+        self, period=PERIOD_OVERALL, limit=None, cacheable=True, stream=False
+    ):
         """Returns the top tracks played by a user.
         * period: The period of time. Possible values:
           o PERIOD_OVERALL
@@ -2498,33 +2534,29 @@ class User(_BaseObject, _Chartable):
         if limit:
             params["limit"] = limit
 
-        return self._get_things("getTopTracks", "track", Track, params, cacheable)
+        return self._get_things("getTopTracks", Track, params, cacheable, stream=stream)
 
-    def get_track_scrobbles(self, artist, track, cacheable=False):
+    def get_track_scrobbles(self, artist, track, cacheable=False, stream=False):
         """
         Get a list of this user's scrobbles of this artist's track,
         including scrobble time.
         """
-
         params = self._get_params()
         params["artist"] = artist
         params["track"] = track
 
-        seq = []
-        for track in _collect_nodes(
-            None, self, self.ws_prefix + ".getTrackScrobbles", cacheable, params
-        ):
-            title = _extract(track, "name")
-            artist = _extract(track, "artist")
-            date = _extract(track, "date")
-            album = _extract(track, "album")
-            timestamp = track.getElementsByTagName("date")[0].getAttribute("uts")
+        def _get_track_scrobbles():
+            for track_node in _collect_nodes(
+                None,
+                self,
+                self.ws_prefix + ".getTrackScrobbles",
+                cacheable,
+                params,
+                stream=stream,
+            ):
+                yield self._extract_played_track(track_node)
 
-            seq.append(
-                PlayedTrack(Track(artist, title, self.network), album, date, timestamp)
-            )
-
-        return seq
+        return _get_track_scrobbles() if stream else list(_get_track_scrobbles())
 
     def get_image(self, size=SIZE_EXTRA_LARGE):
         """
@@ -2569,21 +2601,21 @@ class User(_BaseObject, _Chartable):
 
 class AuthenticatedUser(User):
     def __init__(self, network):
-        User.__init__(self, network.username, network)
+        super().__init__(user_name=network.username, network=network)
 
     def _get_params(self):
         return {"user": self.get_name()}
 
-    def get_name(self):
+    def get_name(self, properly_capitalized=False):
         """Returns the name of the authenticated user."""
-        return self.name
+        return super().get_name(properly_capitalized=properly_capitalized)
 
 
 class _Search(_BaseObject):
     """An abstract class. Use one of its derivatives."""
 
     def __init__(self, ws_prefix, search_terms, network):
-        _BaseObject.__init__(self, network, ws_prefix)
+        super().__init__(network, ws_prefix)
 
         self._ws_prefix = ws_prefix
         self.search_terms = search_terms
@@ -2623,8 +2655,9 @@ class AlbumSearch(_Search):
     """Search for an album by name."""
 
     def __init__(self, album_name, network):
-
-        _Search.__init__(self, "album", {"album": album_name}, network)
+        super().__init__(
+            ws_prefix="album", search_terms={"album": album_name}, network=network
+        )
 
     def get_next_page(self):
         """Returns the next page of results as a sequence of Album objects."""
@@ -2649,7 +2682,9 @@ class ArtistSearch(_Search):
     """Search for an artist by artist name."""
 
     def __init__(self, artist_name, network):
-        _Search.__init__(self, "artist", {"artist": artist_name}, network)
+        super().__init__(
+            ws_prefix="artist", search_terms={"artist": artist_name}, network=network
+        )
 
     def get_next_page(self):
         """Returns the next page of results as a sequence of Artist objects."""
@@ -2676,9 +2711,10 @@ class TrackSearch(_Search):
     """
 
     def __init__(self, artist_name, track_title, network):
-
-        _Search.__init__(
-            self, "track", {"track": track_title, "artist": artist_name}, network
+        super().__init__(
+            ws_prefix="track",
+            search_terms={"track": track_title, "artist": artist_name},
+            network=network,
         )
 
     def get_next_page(self):
@@ -2734,59 +2770,59 @@ def cleanup_nodes(doc):
     return doc
 
 
-def _collect_nodes(limit, sender, method_name, cacheable, params=None):
+def _collect_nodes(limit, sender, method_name, cacheable, params=None, stream=False):
     """
     Returns a sequence of dom.Node objects about as close to limit as possible
     """
-
     if not params:
         params = sender._get_params()
 
-    nodes = []
-    page = 1
-    end_of_pages = False
+    def _stream_collect_nodes():
+        node_count = 0
+        page = 1
+        end_of_pages = False
 
-    while not end_of_pages and (not limit or (limit and len(nodes) < limit)):
-        params["page"] = str(page)
+        while not end_of_pages and (not limit or (limit and node_count < limit)):
+            params["page"] = str(page)
 
-        tries = 1
-        while True:
-            try:
-                doc = sender._request(method_name, cacheable, params)
-                break  # success
-            except Exception as e:
-                if tries >= 3:
-                    raise e
-                # Wait and try again
-                time.sleep(1)
-                tries += 1
+            tries = 1
+            while True:
+                try:
+                    doc = sender._request(method_name, cacheable, params)
+                    break  # success
+                except Exception as e:
+                    if tries >= 3:
+                        raise PyLastError() from e
+                    # Wait and try again
+                    time.sleep(1)
+                    tries += 1
 
-        doc = cleanup_nodes(doc)
+            doc = cleanup_nodes(doc)
 
-        # break if there are no child nodes
-        if not doc.documentElement.childNodes:
-            break
-        main = doc.documentElement.childNodes[0]
+            # break if there are no child nodes
+            if not doc.documentElement.childNodes:
+                break
+            main = doc.documentElement.childNodes[0]
 
-        if main.hasAttribute("totalPages"):
-            total_pages = _number(main.getAttribute("totalPages"))
-        elif main.hasAttribute("totalpages"):
-            total_pages = _number(main.getAttribute("totalpages"))
-        else:
-            raise Exception("No total pages attribute")
+            if main.hasAttribute("totalPages") or main.hasAttribute("totalpages"):
+                total_pages = _number(
+                    main.getAttribute("totalPages") or main.getAttribute("totalpages")
+                )
+            else:
+                raise PyLastError("No total pages attribute")
 
-        for node in main.childNodes:
-            if not node.nodeType == xml.dom.Node.TEXT_NODE and (
-                not limit or (len(nodes) < limit)
-            ):
-                nodes.append(node)
+            for node in main.childNodes:
+                if not node.nodeType == xml.dom.Node.TEXT_NODE and (
+                    not limit or (node_count < limit)
+                ):
+                    node_count += 1
+                    yield node
 
-        if page >= total_pages:
-            end_of_pages = True
+            end_of_pages = page >= total_pages
 
-        page += 1
+            page += 1
 
-    return nodes
+    return _stream_collect_nodes() if stream else list(_stream_collect_nodes())
 
 
 def _extract(node, name, index=0):
@@ -2880,8 +2916,6 @@ def _number(string):
 
     if not string:
         return 0
-    elif string == "":
-        return 0
     else:
         try:
             return int(string)
@@ -2890,9 +2924,6 @@ def _number(string):
 
 
 def _unescape_htmlentity(string):
-
-    # string = _unicode(string)
-
     mapping = html.entities.name2codepoint
     for key in mapping:
         string = string.replace("&%s;" % key, chr(mapping[key]))
